﻿#if ROC_GM_PURCHASE
using System;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using RocketCoroutine;
using RocketGM;
using RocketLog;
using RocketUtils;
using RocketUtils.SerializableDictionary;
using UnityEngine;
using UnityEngine.Purchasing;

namespace RocketGMPurchase
{
    public class RocGMPurchaseController : SingletonBehaviour<RocGMPurchaseController>, IStoreListener
    {
        //[Serializable] private class ProductKeys : SerializableDictionary<string, GMPurchaseProduct> { }

        #region Class Definitions
        public enum GMPurchaseSimulationMode
        {
            None,
            Failed,
            Success
        }

        [Serializable]
        public class LocalProduct
        {
            public ProductDefinition Definition;
            public ProductMetadata Metadata;
            public string Receipt;
            public string TransactionId;
            public int IsNewPayment;
            public GMPurchaseProduct GMPurchaseProduct;
            public Product UnityProduct;
        }

        [Serializable]
        public class GMPurchaseProduct
        {
            public string Id;
            public ProductType Type;
            public string DefaultPrice;
            public bool IsNoAds;
        } 
        #endregion

        private Product[] _unityProductsList;
        private PurchaseFailureReason _purchaseFailureReason;
        private bool _isPurchasing;

        public static event Action<GMPurchaseProduct> PurchaseItemStarted;
        public static event Action<LocalProduct, bool> PurchaseItemSuccessful;
        public static event Action<GMPurchaseProduct, PurchaseFailureReason, bool> PurchaseItemFailed;
        //todo:subscription
        //public static event Action<ProductMetadata, bool> SubscriptionStarted;
        public static event Action RestorationStarted;
        public static event Action RestorationFinished;
        public static event Action ValidationStarted;
        public static event Action ValidationFinished;

        public static event Action StoreProductsListReceived;
        public static event Action<InitializationFailureReason> StoreProductsListReceiveFailed;

        private static IStoreController _mStoreController;          // The Unity Purchasing system.
        private static IExtensionProvider _mStoreExtensionProvider; // The store-specific Purchasing subsystems.

        private Queue<LocalProduct> _waitingPurchaseProductsQueue;
        private LocalProduct _currentProduct;
        private GMPurchaseProduct _tempInappProduct;

        
        [SerializeField] private List<GMPurchaseProduct> _gmPurchaseProducts;
#if ROC_DEBUG_MODE && UNITY_EDITOR
        [Header("DEBUG")]
        [SerializeField] private GMPurchaseSimulationMode _simulationMode;
#else
        private GMPurchaseSimulationMode _simulationMode = GMPurchaseSimulationMode.None;
#endif


        private string _productId;
        private bool _isValidationInProgress;
        private bool _isRestoreInProgress;
        private bool _userInitiated;

        public bool IsInitialized
        {
            get { return _mStoreController != null && _mStoreExtensionProvider != null; }
        }

        protected readonly RocLog Log = new RocLog(typeof(RocGMPurchaseController).Name, DebugLevels.Debug);

#region Initialization	    		
        protected override void OnAwake()
        {
            DontDestroyOnLoad(gameObject);

            Log.Debug("Initialize started", "OnAwake");
            if (IsInitialized)
            {
                // ... we are done here.
                Log.Info("Already initialized", "OnAwake");
                return;
            }

            InternetConnectionController.DoAfterConnectionAvailable(InitializeBuilder);
        }

        private void InitializeBuilder()
        {
            Log.Debug("Initialize builder started", "InitializeBuilder");
            // Create a builder, first passing in a suite of Unity provided stores.
            var builder = ConfigurationBuilder.Instance(StandardPurchasingModule.Instance());

            foreach (GMPurchaseProduct iapProduct in _gmPurchaseProducts)
            {
                builder.AddProduct(iapProduct.Id, iapProduct.Type);
            }

			UnityPurchasing.Initialize(this, builder);
        }

        public void OnInitialized(IStoreController controller, IExtensionProvider extensions)
        {
            _isValidationInProgress = false;
            _waitingPurchaseProductsQueue = new Queue<LocalProduct>();

            // Purchasing has succeeded initializing. Collect our Purchasing references.
            Log.Info("OnInitialized: PASS", "OnInitialized");

            // Overall Purchasing system, configured with products for this application.
            _mStoreController = controller;
            _unityProductsList = _mStoreController.products.all;

            foreach (var product in _unityProductsList)
            {
                Log.Debug("Initialized Product ID: " + product.definition.id, "OnInitialized");
            }
            // Store specific subsystem, for accessing device-specific store features.
            _mStoreExtensionProvider = extensions;

            CoroutineController.StartCoroutine(ValidateWaitingPaymentsQueue());

            if (StoreProductsListReceived != null) StoreProductsListReceived();
        }

        public void OnInitializeFailed(InitializationFailureReason error)
        {
            Log.Info("Billing failed to initialize!", "OnInitializeFailed");

            switch (error)
            {
                case InitializationFailureReason.AppNotKnown:
                    Log.Warning("Is your App correctly uploaded on the relevant publisher console?", "OnInitializeFailed");
                    break;
                case InitializationFailureReason.PurchasingUnavailable:
                    // Ask the user if billing is disabled in device settings.
                    Log.Warning("Billing disabled!", "OnInitializeFailed");
                    break;
                case InitializationFailureReason.NoProductsAvailable:
                    // Developer configuration error; check product metadata.
                    Log.Warning("No products available for purchase!", "OnInitializeFailed");
                    break;
            }

            if (StoreProductsListReceiveFailed != null) StoreProductsListReceiveFailed(error);
        }
#endregion

#region Purchase
        public void BuyItem(string productId)
        {
            if (_isPurchasing)
            {
                Log.Warning("Buying Other Product", "BuyItem");
                return;
            }

            _userInitiated = true;

            _productId = productId;

            if (_unityProductsList != null && _unityProductsList.Length > 0 && Array.Exists(_unityProductsList, product => product.definition.id.Equals(_productId)))
            {
                GMPurchaseProduct gmPurchaseProduct = GetGMPurchaseProduct(productId);

                if (gmPurchaseProduct == null)
                {
                    Log.Error(string.Format("Product with id {0} does NOT exist in GMPurchaseProducts! Please add in prefab!", productId), "BuyItem");
                    return;
                }

                _tempInappProduct = gmPurchaseProduct;

                if (_simulationMode != GMPurchaseSimulationMode.None)
                {
                    SimulatedPurchase(gmPurchaseProduct);
                }
                else
                {
                    BuyProductWithId(gmPurchaseProduct.Id);
                }
            }
            else
            {
                Log.Error(string.Format("Product Type ({0}) does not exist in IapProducts", productId), "BuyItem");
            }
        }

        private void SimulatedPurchase(GMPurchaseProduct product)
        {
            if(PurchaseItemStarted != null)
                PurchaseItemStarted.Invoke(product);

            var boughtProduct = _mStoreController.products.WithID(product.Id);

            _currentProduct = new LocalProduct
            {
                GMPurchaseProduct = product,
                UnityProduct = boughtProduct,
                Definition = boughtProduct.definition,
                IsNewPayment = _userInitiated ? 1 : 0,
                Metadata = boughtProduct.metadata,
                TransactionId = boughtProduct.transactionID,
                Receipt = boughtProduct.receipt
            };

            CoroutineController.DoAfterGivenTime(2f, () =>
            {
                switch (_simulationMode)
                {
                    case GMPurchaseSimulationMode.Success:
                        HandlePurchaseSuccessful();
                        break;

                    case GMPurchaseSimulationMode.Failed:
                        HandlePurchaseFailed(PurchaseFailureReason.PaymentDeclined);
                        break;
                }
            });
        }

        private void BuyProductWithId(string productId)
        {
            if (IsInitialized)
            {
                Product product = _mStoreController.products.WithID(productId);

                if (product != null && product.availableToPurchase)
                {
                    Log.Debug(string.Format("Purchasing product asychronously: '{0}'", product.definition.id), "BuyProductWithId");
                    _isPurchasing = true;
                    _mStoreController.InitiatePurchase(product);

                    if (PurchaseItemStarted != null) PurchaseItemStarted(_tempInappProduct);
                }
                else
                {
                    Log.Error("FAIL! Not purchasing product, either is not found or is not available for purchase", "BuyProductWithId");
                    _userInitiated = false;                    
                    OnPurchaseFailed(productId, PurchaseFailureReason.ProductUnavailable);
                }
            }
            else
            {
                Log.Error("FAIL! Not initialized.", "BuyProductWithId");
                _userInitiated = false;
                OnPurchaseFailed(productId, PurchaseFailureReason.PurchasingUnavailable);
            }
        }

        public PurchaseProcessingResult ProcessPurchase(PurchaseEventArgs args)
        {
            Log.Info("Process Purchase Start", "ProcessPurchase");

            if (!IsInitialized)
            {
                Log.Error("Purchase Controller not initialized yet, wait for initialization!");

                CoroutineController.DoAfterCondition(() => IsInitialized, () => { ProcessPurchase(args); });
                return PurchaseProcessingResult.Pending;
            }

            _isPurchasing = false;
            Product purchaseProduct = args.purchasedProduct;
            GMPurchaseProduct gmPurchaseProduct = GetGMPurchaseProduct(purchaseProduct.definition.id);

            if (gmPurchaseProduct == null)
            {
                Log.Error("Product does NOT exist in IapProducts. Probably a deleted product trying to restore!", "ProcessPurchase");
                return PurchaseProcessingResult.Complete;
            }

            if (!purchaseProduct.hasReceipt || !purchaseProduct.availableToPurchase)
            {
                Log.Warning(string.Format("!purchaseProduct.hasReceipt {0}, !purchaseProduct.availableToPurchase {1}",
                    !purchaseProduct.hasReceipt, !purchaseProduct.availableToPurchase), "ProcessPurchase");
                OnPurchaseFailed(purchaseProduct.definition.id, PurchaseFailureReason.ProductUnavailable);

                return PurchaseProcessingResult.Complete;
            }

            LocalProduct localProduct = new LocalProduct
            {
                Definition = purchaseProduct.definition,
                Metadata = purchaseProduct.metadata,
                Receipt = purchaseProduct.receipt,
                TransactionId = purchaseProduct.transactionID,
                IsNewPayment = _userInitiated ? 1 : 0,
                GMPurchaseProduct = gmPurchaseProduct,
                UnityProduct = purchaseProduct
            };

            Log.Debug("Product Receipt: " + purchaseProduct.receipt, "ProcessPurchase");

            _waitingPurchaseProductsQueue.Enqueue(localProduct);
            Log.Debug("WaitingPurchaseProductsQueue Count >> " + _waitingPurchaseProductsQueue.Count, "ProcessPurchase");

            var purchasingResult = _userInitiated ? PurchaseProcessingResult.Pending : PurchaseProcessingResult.Complete;

            return purchasingResult;
        }
#endregion

#region Validation
        //todo: subscription
        private void ValidateSubscription(string receiptData)
        {
            _currentProduct = new LocalProduct
            {
                Receipt = receiptData
            };

            RocGm.Reporting.Payment(
                true,
                receiptData,
                false,
                callback: OnContinuingSubscriptionValidationCompleted);
        }

        private void OnContinuingSubscriptionValidationCompleted(bool isSuccess, Dictionary<string, object> result)
        {
            Log.Info("Validate Continuing Subscription", "OnContinuingSubscriptionValidationCompleted");

            //todo: subscription
            //var receipt = _currentProduct.Receipt;
            _currentProduct = null;

            //VALIDATION FAILED
            if (!isSuccess)
            {
                Log.Error("Subscription Validation is FAILED!", "OnContinuingSubscriptionValidationCompleted");
                //todo: subscription
                //PlayerData.Instance.SetVipState(false);
                return;
            }

            long expireDate = long.Parse(result["ExpireDate"].ToString());

            Log.Debug("Subscription End Date: " + expireDate, "OnContinuingSubscriptionValidationCompleted");
            Log.Debug("Subscription End Date: " + DateTimeExtensions.FromUnixTimestamp(expireDate), "OnContinuingSubscriptionValidationCompleted");

            //todo: subscription
            //PlayerData.Instance.SetVipState(true, receipt, expireDate);
        }

        IEnumerator ValidateWaitingPaymentsQueue()
        {
            while (true)
            {
                if (RocGm.Reporting.IsPaymentInProgress || _isValidationInProgress || _waitingPurchaseProductsQueue.Count <= 0)
                {
                    yield return new WaitForSecondsRealtime(1f);
                    continue;
                }

                ValidateProduct();
            }
        }

        private void ValidateProduct()
        {
            _isValidationInProgress = true;

            Log.Info("Queue count: " + _waitingPurchaseProductsQueue.Count, "ValidateProduct");

            LocalProduct localProduct = _waitingPurchaseProductsQueue.Dequeue();
            _currentProduct = localProduct;

            _productId = localProduct.GMPurchaseProduct.Id;

            Log.Debug("InProgress Definition >> " + localProduct.Definition.id, "ValidateProduct");
            Log.Debug("InProgress Transaction >> " + localProduct.TransactionId, "ValidateProduct");

            if (ValidationStarted != null) ValidationStarted.Invoke();

            ValidateWaitingPayments(localProduct.Receipt, isSuccess => { _isValidationInProgress = false; });
        }

        private void ValidateWaitingPayments(string receiptData, Action<bool> actionToInvoke = null)
        {
            RocGm.Reporting.Payment(
				_currentProduct.GMPurchaseProduct.Type == ProductType.Subscription,
				receiptData,
				_userInitiated,
				_currentProduct.Metadata.localizedPrice,
				_currentProduct.Metadata.isoCurrencyCode,
				(isSuccess, result) =>
				{
                    OnPurchaseValidationCompleted(isSuccess, result);
					if (actionToInvoke != null) actionToInvoke(isSuccess);
				});

        }

        private void OnPurchaseValidationCompleted(bool isSuccess, Dictionary<string, object> result)
        {
            if (!isSuccess)
            {
                Log.Error("Validation is FAILED. CurrentProduct: " + _currentProduct.TransactionId, "OnPurchaseValidationCompleted");
                OnPurchaseFailed(_currentProduct.Definition.id, PurchaseFailureReason.SignatureInvalid);
                ConfirmPendingPurchase(_currentProduct.UnityProduct);

                _currentProduct = null;
                return;
            }

            // bool isSubscription = _currentProduct.GMPurchaseProduct.Type == ProductType.Subscription;

            //VALIDATION FAILED
            //todo subscription related
            //         if (!isSuccess)
            //{
            //	Log.Error("Validation is FAILED. CurrentProduct: " + _currentProduct.TransactionId);

            //	if (isSubscription)
            //	{
            //		PlayerData.Instance.SetVipState(false);
            //	}

            //	OnPurchaseFailed(_currentProduct.Definition.id, PurchaseFailureReason.SignatureInvalid);
            //	ConfirmPendingPurchase(_currentProduct.UnityProduct);
            //	_isValidationInProgress = false;

            //	return;
            //}

            ////VALIDATION SUCCESSFUL
            //if (isSubscription)
            //{
            //	long expireDate = long.Parse(result["ExpireDate"].ToString());

            //	Log.Warning("Subscription End Date: " + expireDate, "ValidateWaitingPaymentCoroutine");
            //	Log.Warning("Subscription End Date: " + DateTimeExtensions.FromUnixTimestamp(expireDate), "ValidateWaitingPaymentCoroutine");

            //	PlayerData.Instance.SetVipState(true, _currentProduct.Receipt, expireDate);

            //	if (SubscriptionStarted != null) SubscriptionStarted(_currentProduct.Metadata, _userInitiated);
            //}
            //else if (_currentProduct.GMPurchaseProduct.IsNoAds) //NO ADS
            //{
            //	PanPlayerData.IsAdsActive = false;
            //}

            OnInAppProductBought(true, _currentProduct.GMPurchaseProduct);
        }

        private void OnInAppProductBought(bool opResult, GMPurchaseProduct gmPurchaseProduct)
        {
            Log.Debug("OpResult: " + opResult, "OnProductBought");
            Log.Debug("Current Product: " + _currentProduct.UnityProduct.transactionID, "OnProductBought");

            ConfirmPendingPurchase(_currentProduct.UnityProduct);

            if (!opResult || gmPurchaseProduct == null)
            {
                OnPurchaseFailed(_currentProduct.UnityProduct, PurchaseFailureReason.Unknown);
                return;
            }

            HandlePurchaseSuccessful();
        }

        private void ConfirmPendingPurchase(Product product)
        {
            Log.Debug(string.Format("product id: {0}", product.definition.id), "ConfirmPendingPurchase");
            Log.Debug(string.Format("Confirm pending purchases. product hasReceipt: {0}", product.hasReceipt), "ConfirmPendingPurchase");
            Log.Debug(string.Format("Confirm pending purchases. initialized: {0}", IsInitialized), "ConfirmPendingPurchase");

            _isValidationInProgress = false;

            if (!IsInitialized || !product.hasReceipt)
                return;

            Log.Debug("Pending purchase condition passed, confirming pending purchase.", "ConfirmPendingPurchase");
            _mStoreController.ConfirmPendingPurchase(product);
        }
#endregion

        private void OnPurchaseFailed(string productId, PurchaseFailureReason failureReason)
        {
            Log.Error(string.Format("FAIL! Product: '{0}', PurchaseFailureReason: {1}", productId, failureReason), "OnPurchaseFailed");

            HandlePurchaseFailed(failureReason);
        }

        /// <summary>
        /// Unity Callback Function. Should NOT be called explicitly
        /// </summary>
        /// <param name="product"></param>
        /// <param name="failureReason"></param>
        public void OnPurchaseFailed(Product product, PurchaseFailureReason failureReason)
        {
            // Print the failure reason as warning if it is user cancelled.
            if (failureReason == PurchaseFailureReason.UserCancelled)
            {
                Log.Warning(string.Format("FAIL! Product: '{0}', PurchaseFailureReason: {1}", product.definition.storeSpecificId, failureReason), "UnityListenerOnPurchaseFailed");
            }
            else
            {
                Log.Error(string.Format("FAIL! Product: '{0}', PurchaseFailureReason: {1}", product.definition.storeSpecificId, failureReason), "UnityListenerOnPurchaseFailed");
            }
            HandlePurchaseFailed(failureReason);
        }

        private void HandlePurchaseFailed(PurchaseFailureReason failureReason)
        {
            _isPurchasing = false;

            GMPurchaseProduct gmPurchaseProduct = _currentProduct != null ? _currentProduct.GMPurchaseProduct : _tempInappProduct;

            Log.Debug("Waiting Purchase Queue Count: " + _waitingPurchaseProductsQueue.Count, "HandlePurchaseFailed");
            if (_waitingPurchaseProductsQueue.Count == 0 && ValidationFinished != null)
                ValidationFinished.Invoke();

            if (PurchaseItemFailed != null) PurchaseItemFailed(gmPurchaseProduct, failureReason, _userInitiated);

            _userInitiated = false;
        }

        private void HandlePurchaseSuccessful()
        {
            Log.Debug("Waiting Purchase Queue Count: " + _waitingPurchaseProductsQueue.Count, "HandlePurchaseSuccessful");
            if (_waitingPurchaseProductsQueue.Count == 0 && ValidationFinished != null)
                ValidationFinished.Invoke();

            if (PurchaseItemSuccessful != null) PurchaseItemSuccessful(_currentProduct, _userInitiated);

            _userInitiated = false;
            _currentProduct = null;
        }

        /// <summary>
        /// Return Unity Product if exists, otherwise returns NULL
        /// </summary>
        /// <param name="productId"></param>
        /// <returns></returns>
        public Product GetUnityProduct(string productId)
        {
            return _unityProductsList == null ? null : Array.Find(_unityProductsList, x => x.definition.id == productId.ToLower());
        }

        public GMPurchaseProduct GetGMPurchaseProduct(string productId)
        {
            foreach (GMPurchaseProduct gmPurchaseProduct in _gmPurchaseProducts)
            {
                if (gmPurchaseProduct.Id == productId)
                    return gmPurchaseProduct;
            }

            return null;
        }

        public string GetProductPrice(string productId)
        {
            var unityProduct = GetUnityProduct(productId);

            if (unityProduct != null)
            {
                return unityProduct.metadata.localizedPrice.ToString(CultureInfo.InvariantCulture);
            }

            return GetGMPurchaseProduct(productId).DefaultPrice;
        }

        #region Restore
        /// <summary>
        /// USE ONLY FOR IOS!
        /// Restore purchases previously made by this customer. Some platforms automatically restore purchases, like Google.
        /// Apple currently requires explicit purchase restoration for IAP, conditionally displaying a password prompt.
        /// </summary>
        public void RestorePurchases()
        {
            if (_isRestoreInProgress)
                return;

            // If Purchasing has not yet been set up ...
            if (!IsInitialized)
            {
                // ... report the situation and stop restoring. Consider either waiting longer, or retrying initialization.
                Log.Warning("RestorePurchases FAIL. Not initialized.");
                return;
            }

            // If we are running on an Apple device ... 
            if (Application.platform == RuntimePlatform.IPhonePlayer ||
                Application.platform == RuntimePlatform.OSXPlayer)
            {
                if (RestorationStarted != null) RestorationStarted();

                // ... begin restoring purchases
                Log.Info("RestorePurchases started ...");
                _isRestoreInProgress = true;
                // Fetch the Apple store-specific subsystem.
                var apple = _mStoreExtensionProvider.GetExtension<IAppleExtensions>();
                // Begin the asynchronous process of restoring purchases. Expect a confirmation response in 
                // the Action<bool> below, and ProcessPurchase if there are previously purchased products to restore.

                apple.RestoreTransactions((result) =>
                {                    
                    //wait for process purchase
                    //if there is no products to purchase, just close loading
                    CoroutineController.StartCoroutine(WaitForRestoration());
                    // The first phase of restoration. If no more responses are received on ProcessPurchase then 
                    // no purchases are available to be restored.
                    Log.Debug("RestorePurchases continuing: " + result + ". If no further messages, no purchases available to restore.");
                });
            }
            // Otherwise ...
            else
            {
                // We are not running on an Apple device. No work is necessary to restore purchases.
                Log.Error("RestorePurchases FAIL. Not supported on this platform. Current = " + Application.platform);
                OnPurchaseFailed(string.Empty, PurchaseFailureReason.ExistingPurchasePending);
            }
        }

        private IEnumerator WaitForRestoration()
        {            
            yield return new WaitUntil(() => _waitingPurchaseProductsQueue.Count == 0);

            _isRestoreInProgress = false;

            if (RestorationFinished != null) RestorationFinished();
        }
        #endregion

        /// <summary>
        /// Can only be used in DEBUG mode
        /// </summary>
        /// <param name="simulationMode"></param>
        public void ChangeSimulationMode(GMPurchaseSimulationMode simulationMode)
        {
#if ROC_DEBUG_MODE
            _simulationMode = simulationMode;
#endif
        }

    }
}
#endif